<?php
/*!*******************************************************************************
*! FILE NAME  : honeywell_compass.php
*! DESCRIPTION: Reads Heading, Roll, Pitch from the Honeywell HMR3000 compass
*!              module attached to the serial port of the Elphel model 353 camera
*!              Requires 10349/10369 boards with extra connection to +5V
*!              (on 10369 can use 4-pin RJ45 sync out as power).
*!              Needs system to be configured with the debug port disabled.
*! Copyright (C) 2008 Elphel, Inc
*! -----------------------------------------------------------------------------**
*!
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!
*!  $Log: honeywell_compass.php,v $
*!  Revision 1.1.1.1  2008/11/27 20:04:01  elphel
*!
*!
*!  Revision 1.1  2008/04/07 09:17:01  elphel
*!  manual interface to compass modules
*!
*!  Revision 1.1  2008/03/23 20:13:47  elphel
*!  added support for the Honeywell HMR3000 compass module attached to the serial port.
*!
*!
*/


//!test if ttyS0 is not used as console.
  $hpr_cmd=false;
  $error=false;
  $xml = new SimpleXMLElement("<?xml version='1.0'?><hpr/>");
  if (count(file("/proc/tty/driver/serial"))<3) { //no dummy port found - "4: uart:CRISv32 port:00000001 irq:0 tx:0 rx:0 RTS|DTR"
    $xml->addChild ('error',"Serial port 0 is used for debug console, you have to disable it in the system config file to use attached serial devices, like Honeywell HMR3000 Compass module.");
    $error=true;
  } else {
    $s=$_SERVER["QUERY_STRING"]; // starts with "$" or "N" (instead of the "#"), ends with "*"
    if (strlen ($s)<2) {
       $s='$PTNT,HPR*'; //request a single HPR data set
       $hpr_cmd=true;
    } else {
       if ($s[0]=="N") $s[0]="#"; //! just to be able to pass "#" through CGI GET
       if ((($s[0]!='#') && ($s[0]!='$')) || ($s[strlen($s)-1]!='*')) {
         $xml->addChild ('error',"Command requested is '".$s."'. Valid commands start with '$' or '#' (use 'N' instead for CGI url line) and end with '*' (checksum will be added automatically). See HMR3000 manual for the list of commands.");
         $error=true;
       }
    }
  }
  if (!$error) {
    $s1=substr($s,1,strlen($s)-2);
    $cs=0;
    $s1=str_split($s1);
    foreach ($s1 as $c) {
      $cs^=ord($c);
    }
    $s=sprintf("%s%X%X",$s,(($cs>>4) & 0xf),($cs & 0xf));
    exec ("stty -F /dev/ttyS0 onlcr -echo speed 19200");
    $sero=fopen("/dev/ttyS0","w");
    $seri=fopen("/dev/ttyS0","r");
    $n=fwrite($sero,$s."\n");
    $response=fgets($seri);
    fclose ($sero);
    fclose ($seri);
    if ($hpr_cmd) {
      $hpr=explode(",",$response);
      $xml->addChild ('heading',$hpr[1]);
      $xml->addChild ('pitch',$hpr[3]);
      $xml->addChild ('roll',$hpr[5]);
    }
    $xml->addChild ('raw',$response);
  }
  $rslt=$xml->asXML();
  header("Content-Type: text/xml");
  header("Content-Length: ".strlen($rslt)."\n");
  header("Pragma: no-cache\n");
  printf($rslt);
?>
 
