/*
** -----------------------------------------------------------------------------**
** dac2900.v
**
** Interface to a two-channel 10 bit 125MHz DAC (Texas Instruments DAC2900)
** Includes a small LUT RAM where the sequence voltage levels are stored.
**
** Copyright (C) 2003-2006 Elphel, Inc.
**
** -----------------------------------------------------------------------------**
**  This file is part of x324
**  x324 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
**  $Log: dac2900.v,v $
**  Revision 1.1.1.1  2008/11/27 20:04:02  elphel
**
**
**  Revision 1.1.1.1  2007/08/17 12:07:19  elphel
**  This is a fresh tree based on elphel353-2.10
**
**  Revision 1.6  2007/08/17 12:07:19  spectr_rain
**  switch to GPL3 license
**
**  Revision 1.5  2007/07/20 09:46:13  spectr_rain
**  *** empty log message ***
**
**  Revision 1.4  2007/05/25 02:09:09  elphel
**  347 update for mechanical shutter
**
**  Revision 1.1  2006/09/24 16:44:36  elphel
**  Split file into separate module files, added real time clock module, project files
**
*/
module dac2900 (oclk,   //125MHz)
                a,      // [2:0] - output select
                stb,    // synch to clk
                do,     // [9:0]
                wrt1,   //
                wrt2,   //
                ld,     //
// programming interface
                clk,    // write clock, 25MHz
                we,
                addr,    // [3:0] write address, 2*n -> V1, 2*n+1-> V2
                d,
                cswr);

  input        oclk;
  input  [2:0] a;
  input        stb;
  output [9:0] do;
  output       wrt1;
  output       wrt2;
  output       ld;

  input        clk;
  input        we;
  input  [3:0] addr;
  input  [9:0] d;
  input        cswr;

  reg    [3:0] ra;
  wire         wrt0, wrt1, wrt2, ld;
  wire         re=wrt0 || wrt1;
  wire   [9:0] ram_o;
  reg    [9:0] do;

SRL16 i_wrt0 (.Q(wrt0),  .A0(1'b1), .A1(1'b0), .A2(1'b0), .A3(1'b0), .CLK(oclk), .D(stb));   //
SRL16 i_wrt1 (.Q(wrt1),  .A0(1'b1), .A1(1'b0), .A2(1'b0), .A3(1'b0), .CLK(oclk), .D(wrt0));  //
SRL16 i_wrt2 (.Q(wrt2),  .A0(1'b1), .A1(1'b0), .A2(1'b0), .A3(1'b0), .CLK(oclk), .D(wrt1));  //
SRL16 i_ld   (.Q(ld),    .A0(1'b1), .A1(1'b0), .A2(1'b0), .A3(1'b0), .CLK(oclk), .D(wrt2));  //
  always @ (posedge oclk) begin
    if (stb) ra[3:1] <= a[2:0];
    if (stb) ra[0] <=1'b0;
    else if (wrt0) ra[0] <=1'b1;
    if (re) do[9:0] <= ram_o[9:0];
  end

// two ports on DAC have opposite MSB/LSB connections to simplify PCB layout
RAM16XnD #(10) i_ramdac(.D(addr[0]?{d[0],d[1],d[2],d[3],d[4],d[5],d[6],d[7],d[8],d[9]}:d[9:0]),.WE(we && cswr),.CLK(clk),.AW(addr[3:0]),.AR(ra[3:0]),.QR(ram_o[9:0]));
endmodule
