/*
** -----------------------------------------------------------------------------**
** max5438.v
**
** Interface to serial DACs like Maxim MAX5436/max5438, Analog Devices AD7376
**
** Copyright (C) 2003-2006 Elphel, Inc.
**
** -----------------------------------------------------------------------------**
**  This file is part of x324
**  x324 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
**  $Log: max5438.v,v $
**  Revision 1.1.1.1  2008/11/27 20:04:02  elphel
**
**
**  Revision 1.1.1.1  2007/08/17 12:07:19  elphel
**  This is a fresh tree based on elphel353-2.10
**
**  Revision 1.6  2007/08/17 12:07:19  spectr_rain
**  switch to GPL3 license
**
**  Revision 1.5  2007/07/20 09:46:13  spectr_rain
**  *** empty log message ***
**
**  Revision 1.4  2007/05/25 02:09:09  elphel
**  347 update for mechanical shutter
**
**  Revision 1.1  2006/09/24 16:44:36  elphel
**  Split file into separate module files, added real time clock module, project files
**
*/

module max5438 (clk, // 25Mz
                 we,  // 
                 a,   // [1:0] DAC select
                 d,   // [7:0] data (MSB - not used
                 cswr,// decoded address - enable DAC write
                 scl, // serial clock to DACs
                 sda, // serial data to DACs
                 cs); // [3:0] - enable individual DAC, active low
  input clk, we, cswr;
  input [1:0]  a;
  input [7:0]  d;
  output       scl,
               sda;
  output [3:0] cs;

  reg [1:0] sa;
  reg [7:0] sd;
  wire   strt= we && cswr;
  wire   act;  // 
  wire   scl, scl0, frst,last;
  wire [3:0] cs;
  assign cs[3:0]=~{(sa==2'b00) && act,
                   (sa==2'b01) && act,
                   (sa==2'b10) && act,
                   (sa==2'b11) && act};
  assign sda=sd[7];


   FD  i_scl0 (.C(clk), .D (act && !scl0 && !scl), .Q(scl0));
   FD  i_scl  (.C(clk), .D (scl0), .Q(scl));
   FD  i_frst (.C(clk), .D (strt || (frst && !scl)), .Q(frst));
   FD  i_act  (.C(clk), .D (strt || (act && !(scl && last))), .Q(act));
  always @ (posedge clk) begin
    if      (strt) sa[1:0] <= a[1:0];
    if      (strt) sd[7:0] <= d[7:0];
    else if (scl) sd[7:0] <= {sd[6:0],1'b0};
  end
   SRL16E i_last (.Q(last),  .A0(1'b0), .A1(1'b1), .A2(1'b1), .A3(1'b0), .CLK(clk),.CE(scl), .D(frst));  //
 endmodule

