/*
** -----------------------------------------------------------------------------**
** window_proc.v
**
** Manage windowing (individual for each output channel)
**
** Copyright (C) 2003-2006 Elphel, Inc.
**
** -----------------------------------------------------------------------------**
**  This file is part of x324
**  x324 is free software - hardware description language (HDL) code.
** 
**  This program is free software: you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, either version 3 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program.  If not, see <http://www.gnu.org/licenses/>.
** -----------------------------------------------------------------------------**
**
**  $Log: window_proc.v,v $
**  Revision 1.1.1.1  2008/11/27 20:04:02  elphel
**
**
**  Revision 1.1.1.1  2007/08/17 12:07:19  elphel
**  This is a fresh tree based on elphel353-2.10
**
**  Revision 1.6  2007/08/17 12:07:19  spectr_rain
**  switch to GPL3 license
**
**  Revision 1.5  2007/07/20 09:46:13  spectr_rain
**  *** empty log message ***
**
**  Revision 1.4  2007/05/25 02:09:09  elphel
**  347 update for mechanical shutter
**
**  Revision 1.1  2006/09/24 16:44:36  elphel
**  Split file into separate module files, added real time clock module, project files
**
*/


module window_proc (wclk, // interface clock
                we,   // write enable (and cs)
                a,    // [1:0] register address 0..3 - left, width,top, height
                d,    // [11:0]  - 12 bits
                aclk, // ADC/pixel clock
                frame, // high during frame 
                line, // high during pixel shift 
                hact, // valid pixel output
                vact);// active frame output
  input         wclk, we,aclk, frame,line;
  input  [ 1:0] a;
  input  [11:0] d;
  output        hact;
  output        vact;

  reg  line_d, frame_d,vact,vact1,hact,hact0;
  wire line_s= line  && !line_d;
  reg    [11:0] lcnt;
  reg    [11:0] pcnt;
  wire   [11:0] wnd;

  always @ (posedge aclk) begin
    line_d  <= line;
    frame_d <= frame;
    vact  <= frame && (!frame_d || (vact && !((lcnt[11:0]==12'b0) && vact1)));   
    vact1 <= frame && ( vact1 || ((lcnt[11:0]==12'b0) && vact && line_s));
    hact0 <= vact && vact1 && (line_s || (hact0  && !(hact && (pcnt[11:0]==12'b0))));
    hact  <= hact0 && (hact ^ (pcnt[11:0]==12'b0));
    if (!vact || (line_s &&(lcnt[11:0]==12'b0))) lcnt[11:0] <= wnd[11:0];
    else if (line_s) lcnt[11:0] <= lcnt[11:0]-1;

    if (!hact0 || (pcnt[11:0]==12'b0)) pcnt[11:0] <= wnd[11:0];
    else                               pcnt[11:0] <= pcnt[11:0]-1;
  end

  RAM16XnD #(12) i_window(.D(d[11:0]),
                          .WE(we),
                          .CLK(wclk),
                          .AW({2'b00,a[1:0]}),
                          .AR({2'b0,vact1,hact0 || (vact && !vact1)}),
                          .QW(),
                          .QR(wnd[11:0]));
endmodule

