/*!***************************************************************************
*! FILE NAME  : cxsdram353.c
*! DESCRIPTION: TBD
*! Copyright (C) 2002-2008 Elphel, Inc
*! -----------------------------------------------------------------------------**
*!
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!  $Log: cxsdram353.c,v $
*!  Revision 1.1.1.1  2008/11/27 20:04:00  elphel
*!
*!
*!  Revision 1.6  2008/10/25 20:01:33  elphel
*!  just put notes about not using file->f_pos in read() and write()
*!
*!  Revision 1.5  2008/10/06 08:31:08  elphel
*!  snapshot, first images
*!
*!  Revision 1.4  2008/09/20 00:29:49  elphel
*!  moved driver major/minor numbers to a single file - include/asm-cris/elphel/driver_numbers.h
*!
*!  Revision 1.3  2008/09/12 00:23:58  elphel
*!  removed cc353.c, cc353.h
*!
*!  Revision 1.3  2008/08/24 16:16:30  elphel
*!  added support for quantizer tuning (zero bin, shift) and enable writing to the /dev/ccam_img to upload test patterns for compression
*!
*!  Revision 1.2  2008/08/18 06:03:47  elphel
*!  Removed disabling video SDRAM channel 0 (that should be done differently)
*!
*!  Revision 1.1.1.1  2007/08/17 10:23:18  elphel
*!  This is a fresh tree based on elphel353-2.10
*!
*!  Revision 1.7  2007/08/17 10:23:18  spectr_rain
*!  switch to GPL3 license
*!
*!  Revision 1.6  2007/07/20 10:17:45  spectr_rain
*!  *** empty log message ***
*!
*!  Revision 1.1.1.1  2007/02/23 10:11:48  elphel
*!  initial import into CVS
*!
*!  Revision 1.8  2006/09/19 16:22:14  spectr_rain
*!  *** empty log message ***
*!
*!  Revision 1.7  2006/09/02 22:04:17  spectr_rain
*!  *** empty log message ***
*!
*!  Revision 1.6  2006/09/02 00:19:49  spectr_rain
*!  lock sensor while readrawimage
*!
*!  Revision 1.5  2006/09/01 21:36:59  spectr_rain
*!  *** empty log message ***
*!
*!  Revision 1.4  2006/09/01 21:16:00  spectr_rain
*!  *** empty log message ***
*!
*!  Revision 1.3  2006/09/01 21:02:00  spectr_rain
*!  *** empty log message ***
*!
*!  Revision 1.2  2006/09/01 18:54:40  spectr_rain
*!  some fixes for /dev/ccam_img (image.raw)
*!
*!  Revision 1.1.1.1  2006/07/11 19:14:59  spectr_rain
*!  unwork with less than 5MPx Micron sensor, initial branch
*!
*!  Revision 1.2  2005/05/10 21:08:49  elphel
*!  *** empty log message ***
*!

FULL_WIDTH and FULL_HEIGHT increased by 4 each to include 1
*/

/****************** INCLUDE FILES SECTION ***********************************/
#define FULL_WIDTH  ( get_imageParamsThis(P_ACTUAL_WIDTH)+4 )
#define FULL_HEIGHT ( get_imageParamsThis(P_ACTUAL_HEIGHT)+4 )

#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/autoconf.h>

#include <asm/system.h>
//#include <asm/svinto.h>
#include <asm/io.h>

#include <asm/irq.h>

#include <asm/delay.h>
#include <asm/uaccess.h>
#include <asm/elphel/driver_numbers.h>
#include <asm/elphel/c313a.h>

#include "fpgactrl.h"  // defines port_csp0_addr, port_csp4_addr 
#include "framepars.h"  // defines port_csp0_addr, port_csp4_addr 

#include "x3x3.h"
//#include "cc3x3.h"
#include "cxsdram.h"

#define D(x)
#if ELPHEL_DEBUG
  #define MDF(x) {printk("%s:%d:%s ",__FILE__,__LINE__,__FUNCTION__);x ;}
#else
  #define MDF(x)
#endif




#define IMAGERAW_MODULE_DESCRIPTION "Raw image access driver for Elphel 353  camera"
#define	IMAGERAW_DRIVER_NAME        "Raw image access driver"
static const char x353raw_name[] = "elphel_raw353";


static int linebuf_fpn[4096];
static int lineinbuf_fpn;		// should be -1 after frame acquisition, etc - program sensor
static int linebufdirty_fpn = 0;
static int linebuf_img[4096];
static int lineinbuf_img;		// should be -1 after frame acquisition, etc - program sensor
static int linebufdirty_img = 0;

static int minor_lock=0;        /// hardware does not support concurrent accesses, so the driver will be locked for one user


int *   x353raw_fpn_readline(int line, int ignore);
int     x353raw_fpn_writeline(void);
ssize_t x353raw_fpn_read(struct file * file, char * buf, size_t count, loff_t *off);
ssize_t x353raw_fpn_write(struct file * file, const char * buf, size_t count, loff_t *off);
int *   x353raw_frame_readline(int line, int ignore);
int     x353raw_frame_writeline(void);
ssize_t x353raw_frame_read(struct file * file, char * buf, size_t count, loff_t *off);
ssize_t x353raw_frame_write(struct file * file, const char * buf, size_t count, loff_t *off);
loff_t  x353raw_frame_lseek(struct file * file, loff_t offset, int orig);
void    x353raw_frame_release(void); // was missing?
int     x353raw_frame_open(struct inode *inode, struct file *filp);
int     x353raw_fpn_open(struct inode *inode, struct file *filp);
loff_t  x353raw_fpn_lseek(struct file * file, loff_t offset, int orig);
void    x353raw_fpn_release(void);
void   x353raw_f_invalidate (void); // to be called when page number is changed. Write should not be in progress

int     x353raw_open(struct inode *inode, struct file *filp);
int     x353raw_release(struct inode *inode, struct file *filp);
loff_t  x353raw_lseek (struct file * file, loff_t offset, int orig);
ssize_t x353raw_write(struct file * file, const char * buf, size_t count, loff_t *off);
ssize_t x353raw_read(struct file * file, char * buf, size_t count, loff_t *off);





int *x353raw_fpn_readline(int line, int ignore) {	// read fpn line to buffer (returns buffer address). ignore - contents does not matter (to write over)
	int ntilex, padlen, sa, nw, n, i, li;	//nw - number of 32-bit words to read)

	if(line == lineinbuf_fpn)
		return linebuf_fpn;		// cache hit
	if((port_csp0_addr[X313_WA_SD_MODE] & 3) != 3)
		return NULL;			// SDRAM was not enabled or refresh was not on
	if(linebufdirty_fpn) {
		if (x353raw_fpn_writeline() < 0)
			return NULL;		// failed to write buffer
	}
	lineinbuf_fpn = line;
	if(ignore) return linebuf_fpn;
	nw = ((FULL_WIDTH + 1) >> 1);
	ntilex = ((nw + 3) >> 2);
	padlen = ((ntilex + 31) >> 5);
	sa = X313_MAP_FPN + ((padlen * line) << 8);
	X313_INIT_SDCHAN(3, 0, 0, 0, sa, 0x3ff, 0xfff);	// 4096 rows of 32x256 16-bit words each - 64MB > than memory
//	enable channel if it was not already
        X313_CHN_EN(3);
	li = 0;
	while (nw > 0) {
		i = 1000;
		D(printk("wait "));
		while (((i--) > 0) && !X313_SR(PIORDY)) udelay(1);
		D(printk("ready, i=%d\n", (int) i));
		if (i == 0)
			return NULL;		// page not ready
		n = nw;
		if (n > 128)
			n = 128;
		i = RD_SD_PIOWIN;	// will return garbage (because of latency), prepares to the next read, increments address
		for (i = 0; i < n; i++)
			linebuf_fpn[li++] = RD_SD_PIOWIN;
		nw -= n;
		if (nw > 0)
			port_csp0_addr[X313_WA_SDPIO_NEXT] = 0;	// next page (all but last)
	}
	return linebuf_fpn;
}

int x353raw_fpn_writeline(void) {	// write current line of FPN data to SDRAM, clear "dirty" flag
	int ntilex, padlen, sa, nw, n, i, li;

	if(linebufdirty_fpn == 0)
		return 0;
	if(lineinbuf_fpn >= 0) {
		if((port_csp0_addr[X313_WA_SD_MODE] & 3) != 3) return -1;/// SDRAM was not enabled or refresh was not on
		nw = ((FULL_WIDTH + 1) >> 1);
		ntilex = ((nw + 3) >> 2);
		padlen = ((ntilex + 31) >> 5);
		sa = X313_MAP_FPN + ((padlen * lineinbuf_fpn) << 8);
		X313_INIT_SDCHAN(3, 0, 1, 0, sa, 0x3ff, 0xfff);	/// 4096 rows of 32x256 16-bit words each - 64MB > than memory
		/// enable channel if it was not already
    X313_CHN_EN(3);

		li = 0;
		while (nw > 0) {
			i = 1000;
			D(printk("wait "));
			while (((i--) > 0) && !X313_SR(PIORDY)) udelay(1);
			D(printk("ready, i=%d\n", (int) i));
			if (i == 0)  return -1;		/// page not ready
			n = nw;
			if (n > 128)
				n = 128;
			for (i = 0; i < n; i++)
				port_csp0_addr[X313_WA_SD_PIOWIN] = linebuf_fpn[li++];
			nw -= n;
			port_csp0_addr[X313_WA_SDPIO_NEXT] = 0;	// skip 8x32 data (all but last)
		}
		// write the last page if it was partial
//		while(li & 0x3f) {
//			port_csp0_addr[X313_WA_SDPIO_NEXT] = 0; // skip 8x32 data (all but last)
//			li += 8;
//		}
		/// wait page is written
		i = 1000;
		while (((i--) > 0) && !X313_SR(PIOWEMPTY)) udelay(1);
		if (i == 0)
			return -EFAULT;		// page not ready
	}
	linebufdirty_fpn = 0;
	return 0;
}


ssize_t x353raw_fpn_read(struct file * file, char *buf, size_t count, loff_t * off) {
	unsigned long p, fsz, y, xb, xbl, left;
	char *bp = buf;
	char *mp;

	MDF(printk("count=%x, off=%x\n", (int) count, (int) off));
/// ************* NOTE: Never use file->f_pos in write() and read() !!!
        p = *off;
//	printk("x353raw_fpn_read (%x/%x) ", (int) p, (int) count);
	fsz = (FULL_WIDTH * FULL_HEIGHT) << 1;
//	if(p >= fsz)  return -EINVAL; // bigger than SDRAM
	if((p + count) > fsz) {	// truncate count
		count = fsz - p;
	}
	left = count;
	while(left > 0) {
		// calculate start address, pointed by p;
//		printk(":%x:",(int) p);
		y = (p / FULL_WIDTH) >> 1;
//		printk("+%x+",(int) y);
		xb = p - y * (FULL_WIDTH << 1);
		xbl = (FULL_WIDTH << 1) - xb;
		if(xbl > left)
			xbl = left;			// xbl - number of bytes left in the same line
		if(!(mp = (char *)x353raw_fpn_readline(y, 0)))
			return -EFAULT;		// read SDRAM error
//		copy data
		mp += xb;
//		D(printk("copytouser, ibp=%x, nb=%x\n", (int) ibp, (int) nb));
		if(copy_to_user(bp, mp, xbl))
			return -EFAULT;
		left -= xbl;
		p += xbl;
		bp += xbl;
	}							// while (left >0)

   *off+=count;
	return count;
}

ssize_t x353raw_fpn_write(struct file * file, const char *buf, size_t count, loff_t * off) {
	unsigned long p, fsz, y, xb, xbl, left;
	const char *bp = buf;
	char *mp;

//  MDF(printk("x353raw_fpn_write, count=%x, off=%x\n", (int) count, (int) off));
/// ************* NOTE: Never use file->f_pos in write() and read() !!!
   p=*off;
	fsz = (FULL_WIDTH * FULL_HEIGHT) << 1;
	if (p >= fsz)
		return -EINVAL;			// bigger than SDRAM
	if ((p + count) > fsz) {	// truncate count
		count = fsz - p;
	}
	left = count;
	while (left > 0) {
		// calculate start address, pointed by p;
		y = (p / FULL_WIDTH) >> 1;
		xb = p - y * (FULL_WIDTH << 1);
		xbl = (FULL_WIDTH << 1) - xb;
		if (xbl > left)
			xbl = left;			// xbl - number of bytes left in the same line
		// do we need to overwrite a whole line? ((xb==0) && (xbl==(FULL_WIDTH<<1)))
		if (!(mp = (char *) x353raw_fpn_readline(y, ((xb == 0) && (xbl == (FULL_WIDTH << 1))))))
			return -EFAULT;		// read SDRAM error
		// copy to line buffer
		mp += xb;
		if (copy_from_user(mp, bp, xbl))
			return -EFAULT;
		left -= xbl;
		p += xbl;
		bp += xbl;
		linebufdirty_fpn = 1;	// mark as modified
	}							// while (left >0)

       *off+=count;
	return count;
}


loff_t x353raw_fpn_lseek(struct file * file, loff_t offset, int orig) {

	//  orig 0: position from begining
	//  orig 1: relative from current position
	//  orig 2: position from last address


	unsigned long fsz;

	fsz = (FULL_WIDTH * FULL_HEIGHT) << 1;

	MDF(printk("orig= %x, offset=k %x\n",(int) orig, (int) offset));
	switch (orig) {
	case 0:
		file->f_pos = offset;
		break;
	case 1:
		file->f_pos += offset;
		break;
	case 2:
		file->f_pos = fsz + offset;
		break;
	default:
		return -EINVAL;
	}

	// truncate position
	if (file->f_pos < 0) {
		file->f_pos = 0;
		return (-EOVERFLOW);
	}

	if (file->f_pos > fsz) {
		file->f_pos = fsz;
		return (-EOVERFLOW);
	}
	return (file->f_pos);
}


// ----------------------------------------
int *x353raw_frame_readline(int line, int ignore) {	/// read image line to buffer (returns buffer address)- may be modified to use DMA. Also - cache multiple lines? (for color)
	int ntilex, padlen, sa, nw, n, i, li;

//	int line2;
//	int buf128[128];
//	int d;
//	int d, pw;
//	unsigned char *cbuf128 = (unsigned char *) buf128;
//	unsigned char *clinebuf = (unsigned char *) linebuf_img;
//	unsigned short *sbuf128 = (unsigned short *) buf128;
//	unsigned short *slinebuf = (unsigned short *) linebuf_img;

	if(line == lineinbuf_img)
		return linebuf_img;		// cache hit
	if((port_csp0_addr[X313_WA_SD_MODE] & 3) != 3)
		return NULL;			// SDRAM was not enabled or refresh was not on
	if(linebufdirty_img) {
		if (x353raw_frame_writeline() < 0)
			return NULL;		// failed to write buffer
	}
	lineinbuf_img = line;
	if(ignore) return linebuf_img;
	if((get_imageParamsThis(P_BITS) == 8) && !get_imageParamsThis(P_BGFRAME))	// n - number of 32-bit words to read 
		nw = ((FULL_WIDTH + 3) >> 2);
	else
		nw = ((FULL_WIDTH + 1) >> 1);
	ntilex = ((nw + 3) >> 2);
	padlen = ((ntilex + 31) >> 5);

// in external (or virtual) trigger mode always 2 frames acquired (exactly), so there is no need to find the boarder
// just add 2 frames.
// so read line into buffer in any case

	sa = X313_MAP_FRAME + ((padlen * line) << 8);

//pw = FULL_WIDTH;
	X313_INIT_SDCHAN(3, 0, 0, 0, sa, 0x3ff, 0xfff);	// 4096 rows of 32x256 16-bit words each - 64MB > than memory
//printk("readline(): line1 == %d, sa == 0x%08X; nw == %d, padlen == %d, FULL_WIDTH == %d\r\n", line, sa, nw, padlen, pw);

	// enable channel if it was not already
//	port_csp0_addr[X313_WA_SD_MODE] |= 0x23;	// channel3 enabled, refresh, sdram
    X313_CHN_EN(3);

	li = 0;
	while(nw > 0) {
		i = 1000;
		D(printk("wait "));
		while(((i--) > 0) && !X313_SR(PIORDY)) udelay(1);
		D(printk("ready, i = %d\n", (int) i));
		if(i == 0)
			return NULL;		// page not ready
		n = nw;
///TODO: seems 128 can be increased to 256 (and defined in x353.h) - is it a FPGA SDRAM page size?
		if(n > 128) n = 128;
///NOTE:port_csp4_addr can now be replaced by port_csp0_addr to make it faster?
		i = RD_SD_PIOWIN;	// will return garbage (because of latency), prepares to the next read, increments address
		for(i = 0; i < n; i++)
//			linebuf_img[li++] = 0x13141516;
			linebuf_img[li++] = RD_SD_PIOWIN;
		nw -= n;
		if(nw > 0)
			port_csp0_addr[X313_WA_SDPIO_NEXT] = 0;	// next page (all but last)
	}

	// now add line from the second frame if external (or virtual) trigger
/*
	if(get_imageParamsThis(P_TRIG) > 0) {	//(external)
		line2 = line + FULL_HEIGHT;	// second frame
//		sa = X313_MAP_FRAME + ((padlen * line2) << 7);
		sa = X313_MAP_FRAME + ((padlen * line2) << 8);
//		X313_INIT_SDCHAN(3, 0, 0, 0, sa, 0xff, 0x7ff);   // 2048 rows of 16x128 16-bit words each
		X313_INIT_SDCHAN(3, 0, 0, 0, sa, 0x3ff, 0xfff);	// 4096 rows of 32x256 16-bit words each - 64MB > than memory
printk("readline(): line2 == %d, sa == 0x%08X\r\n", line, sa);
		// enable channel if it was not already
//		port_csp0_addr[X313_WA_SD_MODE] |= 0x23;	// channel3 enabled, refresh, sdram
    X313_CHN_EN(3);


		li = 0;
		while(nw > 0) {
			i = 1000;
			while(((i--) > 0) && !X313_SR(PIORDY)) udelay(1);
			if(i == 0)
				return NULL;	// page not ready
			n = nw;
			if(n > 128)
				n = 128;
//			i = RD_SD_PIOWIN;	// will return garbage (because of latency), prepares to the next read, increments address
			for(i = 0; i < n; i++)
				buf128[i] = port_csp0_addr[X313_WA_SD_PIOWIN];
			if((get_imageParamsThis(P_BITS) == 8) && !get_imageParamsThis(P_BGFRAME)) {
				for(i = 0; i < 512; i++) {
					d = clinebuf[li];
					d += cbuf128[i];
					d -= get_imageParamsThis(P_FATZERO);
					if (d < 0)
						d = 0;
					else if (d > 255)
						d = 255;
					clinebuf[li++] = d;
				}
			} else {
				for(i = 0; i < 256; i++) {
					d = slinebuf[li];
					d += sbuf128[i];
					d -= get_imageParamsThis(P_FATZERO);
					if(d < 0)
						d = 0;
					else if(d > 65535)
						d = 65535;
					slinebuf[li++] = d;
				}
			}
			nw -= n;
			if(nw > 0)
				port_csp0_addr[X313_WA_SDPIO_NEXT] = 0;	// next page (all but last)
		}
	}
*/
	return linebuf_img;
}

int x353raw_frame_writeline(void) {	/// write current line of image data to SDRAM, clear "dirty" flag
  int ntilex, padlen, sa, nw, n, i, li;
  if(linebufdirty_img == 0) {
       MDF(printk("lineinbuf_img=0x%x, linebufdirty_img=0x%x\n", lineinbuf_img, linebufdirty_img));
       return 0;
  }
  if(lineinbuf_img >= 0) {
    if((port_csp0_addr[X313_WA_SD_MODE] & 3) != 3) return -1; /// SDRAM was not enabled or refresh was not on
    if((get_imageParamsThis(P_BITS) == 8) && !get_imageParamsThis(P_BGFRAME))  /// nw - number of 32-bit words to read 
      nw = ((FULL_WIDTH + 3) >> 2);
    else
      nw = ((FULL_WIDTH + 1) >> 1);
    ntilex = ((nw + 3) >> 2);
    padlen = ((ntilex + 31) >> 5);
    sa = X313_MAP_FRAME + ((padlen * lineinbuf_img) << 8);

    i=1000;
    while (((i--)>0) && !X313_SR(PIOWEMPTY)) udelay(1); /// We need to make sure write cache is empty before reprogramming channel 3
    if (i==0) return -EFAULT;  /// page not ready

    X313_INIT_SDCHAN(3, 0, 1, 0, sa, 0x3ff, 0xfff); /// 4096 rows of 32x256 16-bit words each - 64MB > than memory
    MDF(printk("lineinbuf_img=0x%x, sa=0x%x, nw=0x%x\n", lineinbuf_img, sa, nw));
    X313_CHN_EN(3);

    li = 0;
    while (nw > 0) {
      i = 1000;
      MDF(printk("wait, nw=0x%x : ", nw));
      while (((i--) > 0) && !X313_SR(PIORDY))  udelay(1);
      MDF(printk("ready, i=%d\n", (int) i));
      if (i == 0) return -EFAULT;		/// page not ready
      n = nw;
///TODO: seems 128 can be increased to 256 (and defined in x353.h) - is it a FPGA SDRAM page size?
      if (n > 128)  n = 128;
      for (i = 0; i < n; i++) port_csp0_addr[X313_WA_SD_PIOWIN] = linebuf_img[li++];
      nw -= n;
      port_csp0_addr[X313_WA_SDPIO_NEXT] = 0; /// skip 8x32 data (all but last)
    }
    /// write the last page if it was partial
    /// wait page is written
    i = 1000;
    while (((i--) > 0) && !X313_SR(PIOWEMPTY))  udelay(1);
    if (i == 0)
    return -EFAULT;		/// page not ready
  }
  linebufdirty_img = 0;
  return 0;
}


ssize_t x353raw_frame_read(struct file * file, char *buf, size_t count, loff_t * off) {
	unsigned long p, fsz, y, xb, xbl, left;
	char *bp = buf;
	char *mp;
        p=*off;
        MDF(printk("count = %x, off = %x, f_pos = %d\r\n", (int)count, (int)off, (int) p));
	fsz = (FULL_WIDTH * FULL_HEIGHT) << (((get_imageParamsThis(P_BITS) == 8) && !get_imageParamsThis(P_BGFRAME)) ? 0 : 1);
        MDF(printk("FULL_WIDTH = 0x%x, FULL_HEIGHT = 0x%x, fsz = 0x%x\r\n", (int) FULL_WIDTH, (int)FULL_HEIGHT, (int) fsz));
	if((p + count) > fsz) {	// truncate count
		count = fsz - p;
	}
	left = count;
	while(left > 0) {
		// calculate start address, pointed by p;
		if((get_imageParamsThis(P_BITS) == 8) && !get_imageParamsThis(P_BGFRAME)) {
			y = p / FULL_WIDTH;
			xb = p - y * FULL_WIDTH;
			xbl = FULL_WIDTH - xb;
		} else {
			y = (p / FULL_WIDTH) >> 1;
			xb = p - y * (FULL_WIDTH << 1);
			xbl = (FULL_WIDTH << 1) - xb;
		}
		if(xbl > left)
			xbl = left;			// xbl - number of bytes left in the same line
		if(!(mp = (char *)x353raw_frame_readline(y + FULL_HEIGHT * get_imageParamsThis(P_PAGE_READ), 0))) {
printk("read SDRAM error!\r\n");
			return -EFAULT;		// read SDRAM error
		}
// copy data
		mp += xb;
//    D(printk ("copytouser, ibp=%x, nb=%x\n", (int) ibp, (int) nb));
		if(copy_to_user(bp, mp, xbl)) {
printk("copy_to_user error!\r\n");
			return -EFAULT;
		}
		left -= xbl;
		p += xbl;
		bp += xbl;
	}							// while (left >0)

   *off+=count;
//printk("count == %d\r\n", count);
	return count;
}

ssize_t x353raw_frame_write(struct file * file, const char * buf, size_t count, loff_t *off) {
   unsigned long p, fsz, y, xb, xbl, left, bytes_in_line;
   const char *bp = buf;
   char *mp;
   p=*off;
//   fsz = (FULL_WIDTH * FULL_HEIGHT) << 1;
   fsz = (FULL_WIDTH * FULL_HEIGHT) << (((get_imageParamsThis(P_BITS) == 8) && !get_imageParamsThis(P_BGFRAME)) ? 0 : 1);

   if (p >= fsz) return -EINVAL; /// bigger than SDRAM
   if ((p + count) > fsz) {	/// truncate count
      count = fsz - p;
   }
   left = count;
   MDF(printk("count == %x, off == %x, f_pos == %d\r\n", (int)count, (int)off, (int) p));
   while (left > 0) {
      /// calculate start address, pointed by p;
      if((get_imageParamsThis(P_BITS) == 8) && !get_imageParamsThis(P_BGFRAME)) {
         y =             p / FULL_WIDTH;
         xb =            p - y * FULL_WIDTH;
         bytes_in_line = FULL_WIDTH;
         xbl =           bytes_in_line - xb;
      } else {
         y =             (p / FULL_WIDTH) >> 1;
         xb =            p - y * (FULL_WIDTH << 1);
         bytes_in_line = (FULL_WIDTH <<1);
         xbl =           bytes_in_line - xb;
      }
      if (xbl > left) xbl = left; /// xbl - number of bytes left in the same line

      /// do we need to overwrite a whole line? ((xb==0) && (xbl==(FULL_WIDTH<<1)))
      if (!(mp = (char *) x353raw_frame_readline(y, ((xb == 0) && (xbl == bytes_in_line))))) return -EFAULT; /// read SDRAM error
      /// copy to line buffer
      mp += xb;
      if (copy_from_user(mp, bp, xbl)) return -EFAULT;
      left -= xbl;
      p += xbl;
      bp += xbl;
      linebufdirty_img = 1; /// mark as modified
   } /// while (left >0)
   *off+=count;
   return count;
}




loff_t x353raw_frame_lseek(struct file * file, loff_t offset, int orig) {
/*
 *  orig 0: position from begning of eeprom
 *  orig 1: relative from current position
 *  orig 2: position from last eeprom address
 */

	unsigned long fsz;
	fsz = (FULL_WIDTH * FULL_HEIGHT) << (((get_imageParamsThis(P_BITS) == 8) && !get_imageParamsThis(P_BGFRAME)) ? 0 : 1);

//	MDF(printk("frame_lseek %x\n", (int) offset));

	switch (orig) {
	case 0:
		file->f_pos = offset;
		break;
	case 1:
		file->f_pos += offset;
		break;
	case 2:
		file->f_pos = fsz + offset;
		break;
	default:
		return -EINVAL;
	}

	/* truncate position */
	if (file->f_pos < 0) {
		file->f_pos = 0;
		return (-EOVERFLOW);
	}

	if (file->f_pos > fsz) {
		file->f_pos = fsz;
		return (-EOVERFLOW);
	}
	return (file->f_pos);
}

int x353raw_frame_open(struct inode *inode, struct file *filp) {	// set filesize
	inode->i_size = (FULL_WIDTH * FULL_HEIGHT) << (((get_imageParamsThis(P_BITS) == 8) && !get_imageParamsThis(P_BGFRAME)) ? 0 : 1);
	linebufdirty_fpn = 0;
	lineinbuf_img = -1;
//	MDF(printk("frame_open, size = %x\n", (int) inode->i_size));
//	printk("frame_open, size = 0x%08X; %d\n", (int)inode->i_size, (int)inode->i_size);
	return 0;
}

void x353raw_frame_release(void) {
  if(linebufdirty_img)  x353raw_frame_writeline();
  MDF(printk("\n"));

}

int x353raw_fpn_open(struct inode *inode, struct file *filp) {	// set filesize
	inode->i_size = (FULL_WIDTH * FULL_HEIGHT) << 1;
	linebufdirty_fpn = 0;
	lineinbuf_fpn = -1;
	MDF(printk("size = %x\n", (int) inode->i_size));
	return 0;
}

void x353raw_fpn_release(void) {
	if(linebufdirty_fpn)
		x353raw_fpn_writeline();
	MDF(printk("\n"));

}

void x353raw_f_invalidate(void) {	// to be called when page number is changed. Write should not be in progress
	linebufdirty_fpn = 0;
	linebufdirty_img = 0;
	lineinbuf_fpn = -1;
	lineinbuf_img = -1;
}

/// File operations - top


static struct file_operations x353raw_fops = {
   owner:    THIS_MODULE,
   llseek:   x353raw_lseek,
   read:     x353raw_read,
   write:    x353raw_write,
   open:     x353raw_open,
   release:  x353raw_release
};

static int __init x353raw_init(void) {
   int res;
   res = register_chrdev(IMAGERAW_MAJOR, x353raw_name, &x353raw_fops);
   if(res < 0) {
     printk(KERN_ERR "x353raw_init: couldn't get a major number %d.\n",IMAGERAW_MAJOR);
     return res;
   }
    printk(IMAGERAW_DRIVER_NAME"- %d\n", IMAGERAW_MAJOR);
    return 0;
}

//#define IMAGERAW_MINOR_FRAME	1
//#define IMAGERAW_MINOR_FPN	2
//#define IMAGERAW_MINOR_UNLOCK	3
//static int minor_lock=0;        /// hardware does not support concurrent accesses, so the driver will be locked for one user

int x353raw_open(struct inode *inode, struct file *filp) {
  int res=0;
  int p = MINOR(inode->i_rdev);
  switch (p) {
    case IMAGERAW_MINOR_UNLOCK:
      minor_lock=0;
      return 0; 
    case IMAGERAW_MINOR_FRAME:
      if (minor_lock) return -EACCES;
      res=x353raw_frame_open(inode, filp);
      break;
    case IMAGERAW_MINOR_FPN:
      if (minor_lock) return -EACCES;
      res=x353raw_fpn_open(inode, filp);
      break;
    default: return -EINVAL;
  }
  minor_lock=p; /// lock driver, save minor for R/W operations
  return 0;
}

int x353raw_release(struct inode *inode, struct file *filp) {
  int p = MINOR(inode->i_rdev);
  switch (p) {
    case IMAGERAW_MINOR_UNLOCK:
      break; 
    case IMAGERAW_MINOR_FRAME:
      x353raw_frame_release();
      break;
    case IMAGERAW_MINOR_FPN:
      x353raw_fpn_release();
      break;
    default: return -EINVAL;
  }
  minor_lock=0; /// unlock driver
  return 0;
}

loff_t x353raw_lseek (struct file * file, loff_t offset, int orig) {
  switch (minor_lock) {
    case IMAGERAW_MINOR_FRAME: return x353raw_frame_lseek(file, offset, orig);
    case IMAGERAW_MINOR_FPN:   return x353raw_fpn_lseek  (file, offset, orig);
    default:                   return -EINVAL;
  }
}

ssize_t x353raw_write(struct file * file, const char * buf, size_t count, loff_t *off) {
  switch (minor_lock) {
    case IMAGERAW_MINOR_FRAME: return x353raw_frame_write(file, buf, count, off);
    case IMAGERAW_MINOR_FPN:   return x353raw_fpn_write  (file, buf, count, off);
    default:                   return -EINVAL;
  }
}

ssize_t x353raw_read(struct file * file, char * buf, size_t count, loff_t *off) {
  switch (minor_lock) {
    case IMAGERAW_MINOR_FRAME: return x353raw_frame_read(file, buf, count, off);
    case IMAGERAW_MINOR_FPN:   return x353raw_fpn_read  (file, buf, count, off);
    default:                   return -EINVAL;
  }
}

module_init(x353raw_init);
MODULE_LICENSE("GPLv3.0");
MODULE_AUTHOR("Andrey Filippov <andrey@elphel.com>.");
MODULE_DESCRIPTION(IMAGERAW_MODULE_DESCRIPTION);


