#!/usr/local/sbin/php -q
<?php
/*!
*! PHP script
*! FILE NAME  : bdetect.php
*! DESCRIPTION: This programs tries to recognize the "sensor" board FPGA using JTAG
*!              and control of the main FPGA of the SCL and SDA pins.
*!              It changes values of these two pins and reads JTAG boundary scan
*!              of the 'sensor board' FPGA.
*! AUTHOR     : elphel, spectr, dzhimiev
*! Copyright (C) 2008 Elphel, Inc
*! -----------------------------------------------------------------------------**
*!
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!
*!  $Log: bdetect.php,v $
*!  Revision 1.3  2009/02/05 19:57:21  dzhimiev
*!  + creates file "var/log/bdetect" that contains a detected extension board ID
*!
*!
*/

/*
 * This programs tries to recognize the "sensor" board FPGA using JTAG
 * and control of the main FPGA of the SCL and SDA pins.
 * It changes values of these two pins and reads JTAG boundary scan
 * of the 'sensor board' FPGA
 
 */
$bdt_log = "/var/log/bdetect";

function readbscan($filename) {
   $jtag  = fopen($filename, 'r+');
   $ffs=  str_repeat ( chr(0xff), 97);
   $ret=  fwrite($jtag, $ffs);
//   printf ("fwrite returned %d<br/>\n",$ret);
   fseek ($jtag,0);
   $boundary= fread($jtag, 97);
   fclose($jtag);
  return $boundary;
}
function read_fpga_reg($addr) {
   $fpga  = fopen('/dev/fpgaio', 'r+');
   fseek ($fpga,$addr<<2);
   $rd=fread($fpga, 4);
   fclose($fpga);
   $v=unpack('V1',$rd);
   print_r($v[1]);
   return $v[1];
}
function checkSclSda($boards) {

// there were changes in registers bits and "fpcf -c .." doesn't work starting from software version 8.0 

//valid for software version 7.x
//   exec("fpcf -c  f 1"); // SCL=0
//   exec("fpcf -c 10 2"); // SCL_EN=1
//   exec("fpcf -c 11 1"); // SDA=0
//   exec("fpcf -c 12 2"); // SDA_EN=1   

//valid for software version 8.x
   exec("fpcf -w 5f 10000"); //SCL=0
   exec("fpcf -w 5f 40000"); //SDA=0
   
   $test=array();
   $boundary=readbscan('/dev/sfpgabscan');
   foreach ($boards as $key => $value) {
     $test[$key]= (((ord($boundary[$value['sda'] >> 3]) >> (7 -($value['sda'] & 7))) &1) == 0) &&
                  (((ord($boundary[$value['scl'] >> 3]) >> (7 -($value['scl'] & 7))) &1) == 0);
   }
   
//   exec("fpcf -c  f 2"); // SCL=1 //v7.x
   exec("fpcf -w 5f 20000"); //SCL=1 //v8.x

   $boundary=readbscan('/dev/sfpgabscan');
   foreach ($boards as $key => $value) {
     $test[$key] = $test[$key] &&
                     (((ord($boundary[$value['sda'] >> 3]) >> (7 -($value['sda'] & 7))) &1) == 0) &&
                     (((ord($boundary[$value['scl'] >> 3]) >> (7 -($value['scl'] & 7))) &1) == 1);
   }
   
//   exec("fpcf -c  f 1"); // SCL=0 //v7.x
//   exec("fpcf -c 11 2"); // SDA=1

   exec("fpcf -w 5f 10000"); //SCL=0 //v8.x
   exec("fpcf -w 5f 80000"); //SDA=1

   $boundary=readbscan('/dev/sfpgabscan');
   foreach ($boards as $key => $value) {
     $test[$key] = $test[$key] &&
                     (((ord($boundary[$value['sda'] >> 3]) >> (7 -($value['sda'] & 7))) &1) == 1) &&
                     (((ord($boundary[$value['scl'] >> 3]) >> (7 -($value['scl'] & 7))) &1) == 0);
   }
   
//   exec("fpcf -c  f 2"); // SCL=1 //v7.x
   exec("fpcf -w 5f 20000"); //SCL=1 //v8.x

   $boundary=readbscan('/dev/sfpgabscan');

//   exec("fpcf -c 10 1");   //SCL_EN=0 //v7.x   
//   exec("fpcf -c 12 1");   //SDA_EN=0

   exec("fpcf -w 5f 30000"); //SCL_EN=0 //v8.x
   exec("fpcf -w 5f c0000"); //SCL_EN=0
   
   foreach ($boards as $key => $value) {
     $test[$key] = $test[$key] &&
                     (((ord($boundary[$value['sda'] >> 3]) >> (7 -($value['sda'] & 7))) &1) == 1) &&
                     (((ord($boundary[$value['scl'] >> 3]) >> (7 -($value['scl'] & 7))) &1) == 1);
     if($test[$key]) {
       return $value['model'];
     }
   }

   return "";
}   
function save_log($log) {
	global $bdt_log;

	$file = fopen($bdt_log, "w+");
	if($file) {
		fwrite($file, $log);
		fclose($file);
	}
}

$boards=array (
                '0' => array ('model' => '10347', 'scl' =>241,'sda' => 199),  // E4, C1
                '1' => array ('model' => '10359', 'scl' =>280,'sda' => 296)   // H6, J5

);


$rslt=checkSclSda($boards);
echo $rslt;
save_log($rslt);
if ($rslt=="") exit (-1);
exit (0);
//echo "<pre>\n";
//var_dump ($rslt);
//echo "</pre>\n";

?> 
