<?php
/*!*******************************************************************************
*! FILE NAME  : focus.php
*! DESCRIPTION: Controls FPGA focus calculation and visualization
*! Copyright (C) 2008 Elphel, Inc
*! -----------------------------------------------------------------------------**
*!
*!  This program is free software: you can redistribute it and/or modify
*!  it under the terms of the GNU General Public License as published by
*!  the Free Software Foundation, either version 3 of the License, or
*!  (at your option) any later version.
*!
*!  This program is distributed in the hope that it will be useful,
*!  but WITHOUT ANY WARRANTY; without even the implied warranty of
*!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*!  GNU General Public License for more details.
*!
*!  You should have received a copy of the GNU General Public License
*!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*! -----------------------------------------------------------------------------**
*!  $Log: focus.php,v $
*!  Revision 1.2  2009/03/16 02:08:57  elphel
*!  Fixed focus helper script to use relative focusing WOI as implemented in 8.x camera driver
*!
*!  Revision 1.1.1.1  2008/11/27 20:04:03  elphel
*!
*!
*!  Revision 1.3  2008/02/01 06:57:15  elphel
*!  Made focus.html read current camera settings at startup
*!
*!  Revision 1.2  2008/02/01 04:52:10  elphel
*!  Fixed focus visualization code in FPGA and related software
*!
*!  Revision 1.1  2008/01/27 06:26:47  elphel
*!  Added Web Focusing Helper
*!
*/

$focus_rdonly=array(
  "FOCUS_SHOW"=>0,
  "FOCUS_SHOW1"=>0,
  "GFOCUS_VALUE"=>0,
  "RFOCUS_LEFT"=>0,
  "RFOCUS_WIDTH"=>0,
  "RFOCUS_TOP"=>0,
  "RFOCUS_HEIGHT"=>0,
  "FOCUS_FILTER"=>0
);
$filter=array(
 array(0,    0.1,  0.25, 0.5,  1.0,  1.0,  0.5,  0,
       0.1,  0.1,  0.25, 0.5,  1.0,  1.0,  0.5,  0,
       0.25, 0.25, 0.5,  1.0,  1.0,  0.5,  0.25, 0,
       0.5,  0.5,  1.0,  1.0,  0.5,  0.25, 0.15, 0,
       1.0,  1.0,  1.0,  0.5,  0.25, 0.15, 0.1,  0,
       1.0,  1.0,  0.5,  0.25, 0.15, 0.1,  0,    0,
       0.5,  0.5,  0.25, 0.15, 0.1,  0,    0,    0,
       0,    0,    0,    0,    0,    0,    0,    0),

 array(0,    0.1,  0.25, 0.5,  1.0,  1.0,  0.5,  0,
       0,    0.1,  0.25, 0.5,  1.0,  1.0,  0.5,  0,
       0,    0.1,  0.25, 0.5,  1.0,  1.0,  0.5,  0,
       0,    0.1,  0.25, 0.5,  1.0,  1.0,  0.5,  0,
       0,    0.1,  0.25, 0.5,  1.0,  1.0,  0.5,  0,
       0,    0.1,  0.25, 0.5,  1.0,  1.0,  0.5,  0,
       0,    0.1,  0.25, 0.5,  1.0,  1.0,  0.5,  0,
       0,    0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0,    0,    0,    0,    0,    0,
       0.1,  0.1,  0.1,  0.1,  0.1,  0.1,  0.1,  0.1,
       0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25, 0.25,
       0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,
       0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,  0.5,
         0,  0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0.1,  0.3,  0.8,  1.0,  1.0,  0,
       0,    0.1,  0.15, 0.4,  0.9,  1.0,  1.0,  0,
       0.1,  0.15, 0.4,  0.6,  1.0,  1.0,  1.0,  0,
       0.3,  0.4,  0.6,  1.0,  1.0,  1.0,  1.0,  0,
       0.8,  0.9,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       0,    0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0.1,  0.3,  0.8,  1.0,  1.0,  0,
       0,    0,    0.1,  0.3,  0.8,  1.0,  1.0,  0,
       0,    0,    0.1,  0.3,  0.8,  1.0,  1.0,  0,
       0,    0,    0.1,  0.3,  0.8,  1.0,  1.0,  0,
       0,    0,    0.1,  0.3,  0.8,  1.0,  1.0,  0,
       0,    0,    0.1,  0.3,  0.8,  1.0,  1.0,  0,
       0,    0,    0.1,  0.3,  0.8,  1.0,  1.0,  0,
       0,    0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0,    0,    0,    0,    0,    0,
       0,    0,    0,    0,    0,    0,    0,    0,
       0.1,  0.1,  0.1,  0.1,  0.1,  0.1,  0.1,  0.1,
       0.3,  0.3,  0.3,  0.3,  0.3,  0.3,  0.3,  0.3,
       0.8,  0.8,  0.8,  0.8,  0.8,  0.8,  0.8,  0.8,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,
         0,  0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0.2,  0.2,  0.2,  0.2,  0.2,  1.0,  1.0,  0,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       0,    0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0,    0,    0,    0,    0,    0,
       0,    0,    0,    0,    0,    0,    0,    0,
       0,    0,    0,    0,    0,    0,    0,    0,
       0,    0,    0,    0,    0,    0,    0,    0,
       0.2,  0.2,  0.2,  0.2,  0.2,  1.0,  1.0,  0,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       0,    0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0,    0,    0,    0.2,  1.0,  0,
       0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0.2,  1.0,  1.0,  1.0,  0,
       0,    0,    0.2,  1.0,  1.0,  1.0,  1.0,  0,
       0,    0.2,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       0.2,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       0,    0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0,    0,    0,    0.2,  1.0,  0,
       0,    0,    0,    0,    0,    0.2,  1.0,  0,
       0,    0,    0,    0,    0,    0.2,  1.0,  0,
       0,    0,    0,    0,    0,    0.2,  1.0,  0,
       0,    0,    0,    0,    0,    0.2,  1.0,  0,
       0,    0,    0,    0,    0,    0.2,  1.0,  0,
       0,    0,    0,    0,    0,    0.2,  1.0,  0,
       0,    0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0,    0,    0,    0,    0,    0,
       0,    0,    0,    0,    0,    0,    0,    0,
       0,    0,    0,    0,    0,    0,    0,    0,
       0,    0,    0,    0,    0,    0,    0,    0,
       0,    0,    0,    0,    0,    0,    0,    0,
       0.2,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       0,    0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0,    0,    0,    0,    0.2,  0,
       0,    0,    0,    0,    0,    0.2,  1.0,  0,
       0,    0,    0,    0,    0.2,  1.0,  1.0,  0,
       0,    0,    0,    0.2,  1.0,  1.0,  1.0,  0,
       0,    0,    0.2,  1.0,  1.0,  1.0,  1.0,  0,
       0,    0.2,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       0.2,  1.0,  1.0,  1.0,  1.0,  1.0,  1.0,  0,
       0,    0,    0,    0,    0,    0,    0,    0),


 array(0,    0,    0.1,  0.3,  0.8,  1.0,  0.2,  0,
       0,    0.1,  0.15, 0.4,  0.9,  1.0,  0.2,  0,
       0.1,  0.15, 0.4,  0.6,  1.0,  0.8,  0.2,  0,
       0.3,  0.4,  0.6,  1.0,  1.0,  0.5,  0.2,  0,
       0.8,  0.9,  1.0,  1.0,  1.0,  0.5,  0.2,  0,
       1.0,  1.0,  0.8,  0.5,  0.5,  0.2,  0.2,  0,
       0.2,  0.2,  0.2,  0.2,  0.2,  0.2,  0.1,  0,
       0,    0,    0,    0,    0,    0,    0,    0),

 array(0,    0,    0.1,  0.3,  0.8,  1.0,  0.2,  0,
       0,    0,    0.1,  0.3,  0.8,  1.0,  0.2,  0,
       0,    0,    0.1,  0.3,  0.8,  1.0,  0.2,  0,
       0,    0,    0.1,  0.3,  0.8,  1.0,  0.2,  0,
       0,    0,    0.1,  0.3,  0.8,  1.0,  0.2,  0,
       0,    0,    0.1,  0.3,  0.8,  1.0,  0.2,  0,
       0,    0,    0.1,  0.1,  0.1,  1.1,  0.1,  0,
       0,    0,    0,    0,    0,    0,    0,    0)
);
$scale=$_GET["SCALE"]*1.0;
$nowait=false;

if (!(($scale>0) && ($scale<100000))) $scale=4096;
if (count($_GET) > 0 ) {
  $pars=array();
  foreach($_GET as $key=>$value) {
     if ($key=="SCALE") {
     } else if ($key=="TABLE") {
       $fn=$value+0;
       if ($fn<0) $fn=0;
       else if ($fn>14) $fn=14;
       for ($i=0; $i<64; $i++) {
        $filter[$fn][$i]=intval($scale*$filter[$fn][$i]);
        if ($filter[$fn][$i]>65535) $filter[$fn][$i]=65535;
       }
       $binary_tab="";
       for ($i=0; $i<64; $i++) $binary_tab.=pack("L",$filter[$fn][$i]);
       $tables_file  = fopen('/dev/fpga_tables', 'w');
       fseek ($tables_file, 0x2000+256*$fn) ; // start of filter table
       fwrite($tables_file, $binary_tab);
       fclose($tables_file);
       elphel_set_P_value(ELPHEL_FOCUS_FILTER,$fn);
       $pars["FOCUS_FILTER"]=$fn;
     } else if ($key=="NOWAIT") {
       $nowait=true;
     } else {
       $pars[$key]=$value+0;
     }
  }
}
//! Program focus parameters if there any passed
  if (count($pars) > 0 ) {
    elphel_set_P_arr ($pars);
//    elphel_program_sensor (1); //nonstop
  }
//! Wait for the next frame (so the FOCUS_VALUE will be updated)
//!NOTE: compressor should be running (continuously or just refreshed), program will wait for it
  if (!$nowait) elphel_wait_frame();
  $state=elphel_get_P_arr ($focus_rdonly);
  $msg="<?xml version=\"1.0\"?>\n<focus>\n";
  foreach($state as $key=>$value) {
    $msg.=sprintf("<%s>%d</%s>\n",$key,$value,$key);
  }
  $msg.="</focus>\n";
  header("Content-Type: text/xml");
  header("Content-Length: ".strlen($msg)."\n");
  header("Pragma: no-cache\n");
  printf($msg);
?> 
